# pylint: skip-file
"""Module Doc."""
import ctypes
import ctypes.util as _ctypes_util
from abc import ABC
import numpy as np
from numpy.ctypeslib import ndpointer, as_array
from organon.fl.mathematics.linearalgebra import enums
from organon.fl.mathematics.linearalgebra.IBlas import IBlas


so_file = _ctypes_util.find_library('mkl_rt')
MKL_FOUND = False
mkl = None
if so_file is not None:
    mkl = ctypes.cdll.LoadLibrary(so_file)
    MKL_FOUND = True


class MklFunctions(IBlas, ABC):
    """Class Doc"""
    @staticmethod
    def saxpy(n: int, a: np.single, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        def mkl_saxpy_argtypes():
            return [ctypes.c_int,
                    ctypes.c_float,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_saxpy.argtypes = mkl_saxpy_argtypes()

        mkl.cblas_saxpy(ctypes.c_int(n),
                        ctypes.c_float(a),
                        x,
                        ctypes.c_int(incx),
                        y,
                        ctypes.c_int(incy))

    @staticmethod
    def daxpy(n: int, a: np.double, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        def mkl_daxpy_argtypes():
            return [ctypes.c_int,
                    ctypes.c_double,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_daxpy.argtypes = mkl_daxpy_argtypes()

        mkl.cblas_daxpy(ctypes.c_int(n),
                        ctypes.c_double(a),
                        x,
                        ctypes.c_int(incx),
                        y,
                        ctypes.c_int(incy))

    @staticmethod
    def saxpby(n: int, a: np.single, x: np.ndarray, incx: int, b: np.single, y: np.ndarray, incy: int):
        MklFunctions.sscal(n, b, y, incy)
        MklFunctions.saxpy(n, a, x, incx, y, incy)

    @staticmethod
    def daxpby(n: int, a: np.double, x: np.ndarray, incx: int, b, y: np.ndarray, incy: int):
        MklFunctions.dscal(n, b, y, incy)
        MklFunctions.daxpy(n, a, x, incx, y, incy)

    @staticmethod
    def scopy(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        def mkl_scopy_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_scopy.argtypes = mkl_scopy_argtypes()

        mkl.cblas_scopy(ctypes.c_int(n),
                        x,
                        ctypes.c_int(incx),
                        y,
                        ctypes.c_int(incy))

    @staticmethod
    def dcopy(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        def mkl_dcopy_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_dcopy.argtypes = mkl_dcopy_argtypes()

        mkl.cblas_dcopy(ctypes.c_int(n),
                        x,
                        ctypes.c_int(incx),
                        y,
                        ctypes.c_int(incy))

    @staticmethod
    def scopynan(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        y[:n * incy:incy] = x[np.logical_not(np.isnan(x))][:n * incx:incx]

    @staticmethod
    def dcopynan(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        y[:n * incy:incy] = x[np.logical_not(np.isnan(x))][:n * incx:incx]

    @staticmethod
    def sdot(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int) -> np.single:
        def mkl_sdot_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_sdot.argtypes = mkl_sdot_argtypes()
        mkl.cblas_sdot.restype = ctypes.c_float

        return mkl.cblas_sdot(ctypes.c_int(n),
                              x,
                              ctypes.c_int(incx),
                              y,
                              ctypes.c_int(incy))

    @staticmethod
    def ddot(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int) -> np.double:
        def mkl_ddot_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_ddot.argtypes = mkl_ddot_argtypes()
        mkl.cblas_ddot.restype = ctypes.c_double

        return mkl.cblas_ddot(ctypes.c_int(n),
                              x,
                              ctypes.c_int(incx),
                              y,
                              ctypes.c_int(incy))

    @staticmethod
    def snrm2(n: int, x: np.ndarray, incx: int) -> np.single:
        def mkl_snrm2_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_snrm2.argtypes = mkl_snrm2_argtypes()
        mkl.cblas_snrm2.restype = ctypes.c_float

        return mkl.cblas_snrm2(ctypes.c_int(n),
                               x,
                               ctypes.c_int(incx))

    @staticmethod
    def dnrm2(n: int, x: np.ndarray, incx: int) -> np.double:
        def mkl_dnrm2_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_dnrm2.argtypes = mkl_dnrm2_argtypes()
        mkl.cblas_dnrm2.restype = ctypes.c_double

        return mkl.cblas_dnrm2(ctypes.c_int(n),
                               x,
                               ctypes.c_int(incx))

    @staticmethod
    def sscal(n: int, a: np.single, x: np.ndarray, incx: int):
        def mkl_sscal_argtypes():
            return [ctypes.c_int,
                    ctypes.c_float,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_sscal.argtypes = mkl_sscal_argtypes()

        mkl.cblas_sscal(ctypes.c_int(n),
                        ctypes.c_float(a),
                        x,
                        ctypes.c_int(incx))

    @staticmethod
    def dscal(n: int, a: np.double, x: np.ndarray, incx: int):
        def mkl_dscal_argtypes():
            return [ctypes.c_int,
                    ctypes.c_double,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_dscal.argtypes = mkl_dscal_argtypes()

        mkl.cblas_dscal(ctypes.c_int(n),
                        ctypes.c_double(a),
                        x,
                        ctypes.c_int(incx))

    @staticmethod
    def sswap(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        def mkl_sswap_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_sswap.argtypes = mkl_sswap_argtypes()

        mkl.cblas_sswap(ctypes.c_int(n),
                        x,
                        ctypes.c_int(incx),
                        y,
                        ctypes.c_int(incy))

    @staticmethod
    def dswap(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        def mkl_dswap_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_dswap.argtypes = mkl_dswap_argtypes()

        mkl.cblas_dswap(ctypes.c_int(n),
                        x,
                        ctypes.c_int(incx),
                        y,
                        ctypes.c_int(incy))

    @staticmethod
    def isamax(n: int, x: np.ndarray, incx: int) -> int:
        def mkl_isamax_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_isamax.argtypes = mkl_isamax_argtypes()
        mkl.cblas_isamax.restype = ctypes.c_int

        return mkl.cblas_isamax(ctypes.c_int(n),
                                x,
                                ctypes.c_int(incx))

    @staticmethod
    def idamax(n: int, x: np.ndarray, incx: int) -> int:
        def mkl_idamax_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_idamax.argtypes = mkl_idamax_argtypes()
        mkl.cblas_idamax.restype = ctypes.c_int

        return mkl.cblas_idamax(ctypes.c_int(n),
                                x,
                                ctypes.c_int(incx))

    @staticmethod
    def isamin(n: int, x: np.ndarray, incx: int) -> int:
        def mkl_isamin_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_isamin.argtypes = mkl_isamin_argtypes()
        mkl.cblas_isamin.restype = ctypes.c_int

        return mkl.cblas_isamin(ctypes.c_int(n),
                                x,
                                ctypes.c_int(incx))

    @staticmethod
    def idamin(n: int, x: np.ndarray, incx: int) -> int:
        def mkl_idamin_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_int]

        mkl.cblas_idamin.argtypes = mkl_idamin_argtypes()
        mkl.cblas_idamin.restype = ctypes.c_int

        return mkl.cblas_idamin(ctypes.c_int(n),
                                x,
                                ctypes.c_int(incx))

    @staticmethod
    def vsadd(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsadd_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsAdd.argtypes = mkl_vsadd_argtypes()

        mkl.vsAdd(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vdadd(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdadd_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdAdd.argtypes = mkl_vdadd_argtypes()

        mkl.vdAdd(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vssub(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vssub_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsSub.argtypes = mkl_vssub_argtypes()

        mkl.vsSub(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vdsub(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdsub_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdSub.argtypes = mkl_vdsub_argtypes()

        mkl.vdSub(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vssqr(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vssqr_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsSqr.argtypes = mkl_vssqr_argtypes()

        mkl.vsSqr(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vdsqr(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdsqr_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdSqr.argtypes = mkl_vdsqr_argtypes()

        mkl.vdSqr(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vsmul(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsmul_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsMul.argtypes = mkl_vsmul_argtypes()

        mkl.vsMul(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vdmul(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdmul_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdMul.argtypes = mkl_vdmul_argtypes()

        mkl.vdMul(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vsabs(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsabs_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsAbs.argtypes = mkl_vsabs_argtypes()

        mkl.vsAbs(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vdabs(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdabs_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdAbs.argtypes = mkl_vdabs_argtypes()

        mkl.vdAbs(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vsfmod(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsfmod_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsFmod.argtypes = mkl_vsfmod_argtypes()

        mkl.vsFmod(ctypes.c_int(n),
                   a,
                   b,
                   y)

    @staticmethod
    def vdfmod(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdfmod_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdFmod.argtypes = mkl_vdfmod_argtypes()

        mkl.vdFmod(ctypes.c_int(n),
                   a,
                   b,
                   y)

    @staticmethod
    def vsremainder(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsremainder_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsRemainder.argtypes = mkl_vsremainder_argtypes()

        mkl.vsRemainder(ctypes.c_int(n),
                        a,
                        b,
                        y)

    @staticmethod
    def vdremainder(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdremainder_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdRemainder.argtypes = mkl_vdremainder_argtypes()

        mkl.vdRemainder(ctypes.c_int(n),
                        a,
                        b,
                        y)

    @staticmethod
    def vsinv(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsinv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsInv.argtypes = mkl_vsinv_argtypes()

        mkl.vsInv(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vdinv(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdinv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdInv.argtypes = mkl_vdinv_argtypes()

        mkl.vdInv(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vsdiv(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsdiv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsDiv.argtypes = mkl_vsdiv_argtypes()

        mkl.vsDiv(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vddiv(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vddiv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdDiv.argtypes = mkl_vddiv_argtypes()

        mkl.vdDiv(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vssqrt(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vssqrt_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsSqrt.argtypes = mkl_vssqrt_argtypes()

        mkl.vsSqrt(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vdsqrt(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdsqrt_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdSqrt.argtypes = mkl_vdsqrt_argtypes()

        mkl.vdSqrt(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vspow(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vspow_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsPow.argtypes = mkl_vspow_argtypes()

        mkl.vsPow(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vdpow(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdpow_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdPow.argtypes = mkl_vdpow_argtypes()

        mkl.vdPow(ctypes.c_int(n),
                  a,
                  b,
                  y)

    @staticmethod
    def vspowx(n: int, a: np.ndarray, b: np.single, y: np.ndarray):
        def mkl_vspowx_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_float,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsPowx.argtypes = mkl_vspowx_argtypes()

        mkl.vsPowx(ctypes.c_int(n),
                   a,
                   ctypes.c_float(b),
                   y)

    @staticmethod
    def vdpowx(n: int, a: np.ndarray, b: np.single, y: np.ndarray):
        def mkl_vdpowx_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.c_double,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdPowx.argtypes = mkl_vdpowx_argtypes()

        mkl.vdPowx(ctypes.c_int(n),
                   a,
                   ctypes.c_double(b),
                   y)

    @staticmethod
    def vsexp(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsexp_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsExp.argtypes = mkl_vsexp_argtypes()

        mkl.vsExp(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vdexp(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdexp_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdExp.argtypes = mkl_vdexp_argtypes()

        mkl.vdExp(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vsexp2(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsexp2_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsExp2.argtypes = mkl_vsexp2_argtypes()

        mkl.vsExp2(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vdexp2(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdexp2_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdExp2.argtypes = mkl_vdexp2_argtypes()

        mkl.vdExp2(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vsexp10(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsexp10_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsExp10.argtypes = mkl_vsexp10_argtypes()

        mkl.vsExp10(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vdexp10(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdexp10_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdExp10.argtypes = mkl_vdexp10_argtypes()

        mkl.vdExp10(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vsln(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsln_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsLn.argtypes = mkl_vsln_argtypes()

        mkl.vsLn(ctypes.c_int(n),
                 a,
                 y)

    @staticmethod
    def vdln(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdln_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdLn.argtypes = mkl_vdln_argtypes()

        mkl.vdLn(ctypes.c_int(n),
                 a,
                 y)

    @staticmethod
    def vslog2(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vslog2_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsLog2.argtypes = mkl_vslog2_argtypes()

        mkl.vsLog2(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vdlog2(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdlog2_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdLog2.argtypes = mkl_vdlog2_argtypes()

        mkl.vdLog2(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vslog10(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vslog10_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsLog10.argtypes = mkl_vslog10_argtypes()

        mkl.vsLog10(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vdlog10(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdlog10_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdLog10.argtypes = mkl_vdlog10_argtypes()

        mkl.vdLog10(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vserf(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vserf_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsErf.argtypes = mkl_vserf_argtypes()

        mkl.vsErf(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vderf(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vderf_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdErf.argtypes = mkl_vderf_argtypes()

        mkl.vdErf(ctypes.c_int(n),
                  a,
                  y)

    @staticmethod
    def vserfc(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vserfc_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsErfc.argtypes = mkl_vserfc_argtypes()

        mkl.vsErfc(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vderfc(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vderfc_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdErfc.argtypes = mkl_vderfc_argtypes()

        mkl.vdErfc(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vscdfnorm(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vscdfnorm_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsCdfNorm.argtypes = mkl_vscdfnorm_argtypes()

        mkl.vsCdfNorm(ctypes.c_int(n),
                      a,
                      y)

    @staticmethod
    def vdcdfnorm(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdcdfnorm_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdCdfNorm.argtypes = mkl_vdcdfnorm_argtypes()

        mkl.vdCdfNorm(ctypes.c_int(n),
                      a,
                      y)

    @staticmethod
    def vserfinv(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vserfinv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsErfInv.argtypes = mkl_vserfinv_argtypes()

        mkl.vsErfInv(ctypes.c_int(n),
                     a,
                     y)

    @staticmethod
    def vderfinv(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vderfinv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdErfInv.argtypes = mkl_vderfinv_argtypes()

        mkl.vdErfInv(ctypes.c_int(n),
                     a,
                     y)

    @staticmethod
    def vserfcinv(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vserfcinv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsErfcInv.argtypes = mkl_vserfcinv_argtypes()

        mkl.vsErfcInv(ctypes.c_int(n),
                      a,
                      y)

    @staticmethod
    def vderfcinv(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vderfcinv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdErfcInv.argtypes = mkl_vderfcinv_argtypes()

        mkl.vdErfcInv(ctypes.c_int(n),
                      a,
                      y)

    @staticmethod
    def vscdfnorminv(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vscdfnorminv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsCdfNormInv.argtypes = mkl_vscdfnorminv_argtypes()

        mkl.vsCdfNormInv(ctypes.c_int(n),
                         a,
                         y)

    @staticmethod
    def vdcdfnorminv(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdcdfnorminv_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdCdfNormInv.argtypes = mkl_vdcdfnorminv_argtypes()

        mkl.vdCdfNormInv(ctypes.c_int(n),
                         a,
                         y)

    @staticmethod
    def vslgamma(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vslgamma_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsLGamma.argtypes = mkl_vslgamma_argtypes()

        mkl.vsLGamma(ctypes.c_int(n),
                     a,
                     y)

    @staticmethod
    def vdlgamma(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdlgamma_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdLGamma.argtypes = mkl_vdlgamma_argtypes()

        mkl.vdLGamma(ctypes.c_int(n),
                     a,
                     y)

    @staticmethod
    def vstgamma(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vstgamma_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsTGamma.argtypes = mkl_vstgamma_argtypes()

        mkl.vsTGamma(ctypes.c_int(n),
                     a,
                     y)

    @staticmethod
    def vdtgamma(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdtgamma_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdTGamma.argtypes = mkl_vdtgamma_argtypes()

        mkl.vdTGamma(ctypes.c_int(n),
                     a,
                     y)

    @staticmethod
    def vsexpint1(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsexpint1_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsExpInt1.argtypes = mkl_vsexpint1_argtypes()

        mkl.vsExpInt1(ctypes.c_int(n),
                      a,
                      y)

    @staticmethod
    def vdexpint1(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdexpint1_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdExpInt1.argtypes = mkl_vdexpint1_argtypes()

        mkl.vdExpInt1(ctypes.c_int(n),
                      a,
                      y)

    @staticmethod
    def vsfloor(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsfloor_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsFloor.argtypes = mkl_vsfloor_argtypes()

        mkl.vsFloor(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vdfloor(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdfloor_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdFloor.argtypes = mkl_vdfloor_argtypes()

        mkl.vdFloor(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vsceil(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsceil_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsCeil.argtypes = mkl_vsceil_argtypes()

        mkl.vsCeil(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vdceil(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdceil_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdCeil.argtypes = mkl_vdceil_argtypes()

        mkl.vdCeil(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vstrunc(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vstrunc_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsTrunc.argtypes = mkl_vstrunc_argtypes()

        mkl.vsTrunc(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vdtrunc(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdtrunc_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdTrunc.argtypes = mkl_vdtrunc_argtypes()

        mkl.vdTrunc(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vsround(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsround_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsRound.argtypes = mkl_vsround_argtypes()

        mkl.vsRound(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vdround(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdround_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdRound.argtypes = mkl_vdround_argtypes()

        mkl.vdRound(ctypes.c_int(n),
                    a,
                    y)

    @staticmethod
    def vsnearbyint(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsnearbyint_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsNearbyInt.argtypes = mkl_vsnearbyint_argtypes()

        mkl.vsNearbyInt(ctypes.c_int(n),
                        a,
                        y)

    @staticmethod
    def vdnearbyint(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdnearbyint_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdNearbyInt.argtypes = mkl_vdnearbyint_argtypes()

        mkl.vdNearbyInt(ctypes.c_int(n),
                        a,
                        y)

    @staticmethod
    def vsrint(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsrint_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsRint.argtypes = mkl_vsrint_argtypes()

        mkl.vsRint(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vdrint(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdrint_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdRint.argtypes = mkl_vdrint_argtypes()

        mkl.vdRint(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vsmodf(n: int, a: np.ndarray, y: np.ndarray, z: np.ndarray):
        def mkl_vsmodf_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsModf.argtypes = mkl_vsmodf_argtypes()

        mkl.vsModf(ctypes.c_int(n),
                   a,
                   y,
                   z)

    @staticmethod
    def vdmodf(n: int, a: np.ndarray, y: np.ndarray, z: np.ndarray):
        def mkl_vdmodf_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdModf.argtypes = mkl_vdmodf_argtypes()

        mkl.vdModf(ctypes.c_int(n),
                   a,
                   y,
                   z)

    @staticmethod
    def vsfrac(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vsfrac_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsFrac.argtypes = mkl_vsfrac_argtypes()

        mkl.vsFrac(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vdfrac(n: int, a: np.ndarray, y: np.ndarray):
        def mkl_vdfrac_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdFrac.argtypes = mkl_vdfrac_argtypes()

        mkl.vdFrac(ctypes.c_int(n),
                   a,
                   y)

    @staticmethod
    def vsfmax(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsfmax_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsFmax.argtypes = mkl_vsfmax_argtypes()

        mkl.vsFmax(ctypes.c_int(n),
                   a,
                   b,
                   y)

    @staticmethod
    def vdfmax(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdfmax_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdFmax.argtypes = mkl_vdfmax_argtypes()

        mkl.vdFmax(ctypes.c_int(n),
                   a,
                   b,
                   y)

    @staticmethod
    def vsfmin(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsfmin_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsFmin.argtypes = mkl_vsfmin_argtypes()

        mkl.vsFmin(ctypes.c_int(n),
                   a,
                   b,
                   y)

    @staticmethod
    def vdfmin(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdfmin_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdFmin.argtypes = mkl_vdfmin_argtypes()

        mkl.vdFmin(ctypes.c_int(n),
                   a,
                   b,
                   y)

    @staticmethod
    def vsmaxmag(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsmaxmag_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsMaxMag.argtypes = mkl_vsmaxmag_argtypes()

        mkl.vsMaxMag(ctypes.c_int(n),
                     a,
                     b,
                     y)

    @staticmethod
    def vdmaxmag(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdmaxmag_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdMaxMag.argtypes = mkl_vdmaxmag_argtypes()

        mkl.vdMaxMag(ctypes.c_int(n),
                     a,
                     b,
                     y)

    @staticmethod
    def vsminmag(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vsminmag_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_float, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vsMinMag.argtypes = mkl_vsminmag_argtypes()

        mkl.vsMinMag(ctypes.c_int(n),
                     a,
                     b,
                     y)

    @staticmethod
    def vdminmag(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        def mkl_vdminmag_argtypes():
            return [ctypes.c_int,
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_double, ndim=1, flags='C_CONTIGUOUS')]

        mkl.vdMinMag.argtypes = mkl_vdminmag_argtypes()

        mkl.vdMinMag(ctypes.c_int(n),
                     a,
                     b,
                     y)

    # SPARSE FUNCTIONS

    sparse_matrix_t = ctypes.POINTER(enums.SparseMatrix)

    @staticmethod
    def _create_handle(indexing: enums.SparseIndexBaseT, rows: int, cols: int, rows_start: np.ndarray,
                       rows_end: np.ndarray,
                       col_indx: np.ndarray, values: np.ndarray, dtype: np.dtype, sparse_format='csr'):
        c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

        if sparse_format == 'csr':
            create_func = mkl.mkl_sparse_s_create_csr if dtype == np.single else mkl.mkl_sparse_d_create_csr
        else:
            create_func = mkl.mkl_sparse_s_create_csc if dtype == np.single else mkl.mkl_sparse_d_create_csc

        create_func.argtypes = [ctypes.POINTER(MklFunctions.sparse_matrix_t),
                                ctypes.c_int,
                                ctypes.c_long,
                                ctypes.c_long,
                                ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                                ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                                ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                                ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS')]
        create_func.restype = ctypes.c_int
        handle = MklFunctions.sparse_matrix_t()
        create_func(ctypes.byref(handle),
                    ctypes.c_int(indexing.value),
                    ctypes.c_long(rows),
                    ctypes.c_long(cols),
                    rows_start,
                    rows_end,
                    col_indx,
                    values)
        return handle

    @staticmethod
    def _sp_optimize_mm(sp_handle: sparse_matrix_t, operation: enums.SparseOperationT,
                        descr: enums.MatrixDescr, layout: enums.SparseLayoutT,
                        dense_size: int,
                        exp_calls: int):
        mkl.mkl_sparse_set_mm_hint.argtypes = [MklFunctions.sparse_matrix_t,
                                               ctypes.c_int,
                                               enums.MatrixDescr,
                                               ctypes.c_int,
                                               ctypes.c_int,
                                               ctypes.c_int]
        mkl.mkl_sparse_set_mm_hint(sp_handle,
                                   ctypes.c_int(operation.value),
                                   descr,
                                   ctypes.c_int(layout.value),
                                   ctypes.c_int(dense_size),
                                   ctypes.c_int(exp_calls))
        mkl.mkl_sparse_optimize.argtypes = [MklFunctions.sparse_matrix_t]
        mkl.mkl_sparse_optimize(sp_handle)

    @staticmethod
    def _export_sparse(sp_handle: sparse_matrix_t, c, jc, ic, dtype: np.dtype, sparse_format='csr'):
        c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

        data = ctypes.POINTER(c_dtype)()
        indptrb = ctypes.POINTER(ctypes.c_long)()
        indptren = ctypes.POINTER(ctypes.c_long)()
        indices = ctypes.POINTER(ctypes.c_long)()

        ordering = ctypes.c_int()
        nrows = ctypes.c_long()
        ncols = ctypes.c_long()

        if sparse_format == 'csr':
            out_func = mkl.mkl_sparse_s_export_csr if dtype == np.single else mkl.mkl_sparse_d_export_csr
        else:
            out_func = mkl.mkl_sparse_s_export_csc if dtype == np.single else mkl.mkl_sparse_d_export_csc

        out_func(sp_handle,
                 ctypes.byref(ordering),
                 ctypes.byref(nrows),
                 ctypes.byref(ncols),
                 ctypes.byref(indptrb),
                 ctypes.byref(indptren),
                 ctypes.byref(indices),
                 ctypes.byref(data))

        # ncols = ncols.value
        nrows = nrows.value

        indptrb = as_array(indptrb, shape=(nrows,))
        indptren = as_array(indptren, shape=(nrows,))

        ic[:nrows + 1] = np.insert(indptren, 0, indptrb[0])
        nnz = indptren[-1] - indptrb[0]

        c[:nnz] = np.array(as_array(data, shape=(nnz,)), copy=True)
        jc[:nnz] = np.array(as_array(indices, shape=(nnz,)), copy=True)

        # return sparse.csr_matrix((data, indices, indptren), shape=(nrows, ncols))

    @staticmethod
    def sparse_s_mv(operation: enums.SparseOperationT, m: int, k: int, alpha: np.single,
                    descr: enums.MatrixDescr, val: np.ndarray,
                    indx: np.ndarray,
                    pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, beta: np.single,
                    y: np.ndarray, sparse_format='csr'):
        MklFunctions._sparse_mv(operation, m, k, alpha, descr, val, indx, pntrb, pntre, x, beta, y, np.single,
                                sparse_format)

    @staticmethod
    def sparse_d_mv(operation: enums.SparseOperationT, m: int, k: int, alpha: np.double,
                    descr: enums.MatrixDescr, val: np.ndarray,
                    indx: np.ndarray,
                    pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, beta: np.double,
                    y: np.ndarray, sparse_format='csr'):
        MklFunctions._sparse_mv(operation, m, k, alpha, descr, val, indx, pntrb, pntre, x, beta, y, np.double,
                                sparse_format)

    @staticmethod
    # Sparse matrix dense vector yeni
    def _sparse_mv(operation, m, k, alpha, descr, val, indx, pntrb, pntre, x, beta, y, dtype, sparse_format='csr'):
        c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

        ref = MklFunctions._create_handle(enums.SparseIndexBaseT.SPARSE_INDEX_BASE_ZERO,
                                          m,
                                          k,
                                          pntrb,
                                          pntre,
                                          indx,
                                          val,
                                          dtype,
                                          sparse_format)

        mkl.mkl_sparse_optimize.argtypes = [MklFunctions.sparse_matrix_t]
        mkl.mkl_sparse_optimize(ref)

        sparse_mv = mkl.mkl_sparse_s_mv if dtype == np.single else mkl.mkl_sparse_d_mv

        sparse_mv.argtypes = [ctypes.c_int,
                              c_dtype,
                              MklFunctions.sparse_matrix_t,
                              enums.MatrixDescr,
                              ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                              c_dtype,
                              ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS')]

        sparse_mv(ctypes.c_int(operation.value),
                  c_dtype(alpha),
                  ref,
                  descr,
                  x,
                  c_dtype(beta),
                  y)

        mkl.mkl_sparse_destroy(ref)

    @staticmethod
    def sparse_s_mm(operation: enums.SparseOperationT, m: int, n: int, k: int, alpha: np.single,
                    descr: enums.MatrixDescr, layout: enums.SparseLayoutT, val: np.ndarray,
                    indx: np.ndarray, pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, ldx: int, beta: np.single,
                    y: np.ndarray, ldy: int):
        _sparse_mm(operation, m, n, k, alpha, descr, val, indx, pntrb, pntre, layout, x, ldx, beta, y, ldy,
                   np.single)

    @staticmethod
    def sparse_d_mm(operation: enums.SparseOperationT, m: int, n: int, k: int, alpha: np.double,
                    descr: enums.MatrixDescr, layout: enums.SparseLayoutT, val: np.ndarray,
                    indx: np.ndarray, pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, ldx: int, beta: np.double,
                    y: np.ndarray, ldy: int):
        _sparse_mm(operation, m, n, k, alpha, descr, val, indx, pntrb, pntre, layout, x, ldx, beta, y, ldy,
                   np.double)

    @staticmethod
    def sparse_s_spmm(operation, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info):
        return MklFunctions._sparse_spmm(operation, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax,
                                         info, np.single)

    @staticmethod
    def sparse_d_spmm(operation, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info):
        return MklFunctions._sparse_spmm(operation, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax,
                                         info, np.double)

    @staticmethod
    # Sparse matrix Sparse matrix yeni
    # trans, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info, dtype
    def _sparse_spmm(operation, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info, dtype):
        refa = MklFunctions._create_handle(enums.SparseIndexBaseT.SPARSE_INDEX_BASE_ZERO,
                                           m,
                                           n,
                                           ia[0:-1],
                                           ia[1:],
                                           ja,
                                           a,
                                           dtype)

        refb = MklFunctions._create_handle(enums.SparseIndexBaseT.SPARSE_INDEX_BASE_ZERO,
                                           n,
                                           k,
                                           ib[0:-1],
                                           ib[1:],
                                           jb,
                                           b,
                                           dtype)

        ref_handle = MklFunctions.sparse_matrix_t()

        mkl.mkl_sparse_spmm(ctypes.c_int(operation.value),
                            refa,
                            refb,
                            ctypes.byref(ref_handle))

        MklFunctions._export_sparse(ref_handle, c, jc, ic, dtype)

    @staticmethod
    def sparse_s_add(operation, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info):
        return MklFunctions._sparse_add(operation, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax,
                                        info, np.single)

    @staticmethod
    def sparse_d_add(operation, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info):
        return MklFunctions._sparse_add(operation, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax,
                                        info, np.double)

    @staticmethod
    # Sparse matrix plus Sparse matrix yeni
    # trans, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info, dtype
    def _sparse_add(operation, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info, dtype):
        c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

        refa = MklFunctions._create_handle(enums.SparseIndexBaseT.SPARSE_INDEX_BASE_ZERO,
                                           m,
                                           n,
                                           ia[0:-1],
                                           ia[1:],
                                           ja,
                                           a,
                                           dtype)

        refb = MklFunctions._create_handle(enums.SparseIndexBaseT.SPARSE_INDEX_BASE_ZERO,
                                           m,
                                           n,
                                           ib[0:-1],
                                           ib[1:],
                                           jb,
                                           b,
                                           dtype)

        ref_handle = MklFunctions.sparse_matrix_t()

        handle = mkl.mkl_sparse_s_add if dtype == np.single else mkl.mkl_sparse_d_add

        handle(ctypes.c_int(operation.value),
               refa,
               c_dtype(beta),
               refb,
               ctypes.byref(ref_handle))

        return MklFunctions._export_sparse(ref_handle, c, jc, ic, dtype)

    @staticmethod
    def scsrmv(transa: ctypes.c_char_p, m: int, k: int, alpha: np.single,
               matdescra: np.ndarray, val: np.ndarray, indx: np.ndarray,
               pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, beta: np.single, y: np.ndarray):
        MklFunctions._csrmv(transa, m, k, alpha, matdescra, val, indx, pntrb, pntre, x, beta, y, np.single)

    @staticmethod
    def dcsrmv(transa: ctypes.c_char_p, m: int, k: int, alpha: np.double,
               matdescra: np.ndarray, val: np.ndarray, indx: np.ndarray,
               pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, beta: np.double, y: np.ndarray):
        MklFunctions._csrmv(transa, m, k, alpha, matdescra, val, indx, pntrb, pntre, x, beta, y, np.double)

    @staticmethod
    # Sparse matrix dense vector eski
    def _csrmv(transa, m, k, alpha, matdescra, val, indx, pntrb, pntre, x, beta, y, dtype):
        c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

        def mkl_csrmv_argtypes():
            return [ctypes.c_char_p,
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(c_dtype),
                    ndpointer(dtype=ctypes.c_char, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.POINTER(c_dtype),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS')]

        csrmv = mkl.mkl_scsrmv if dtype == np.single else mkl.mkl_dcsrmv

        csrmv.argtypes = mkl_csrmv_argtypes()

        csrmv(transa,
              ctypes.byref(ctypes.c_int(m)),
              ctypes.byref(ctypes.c_int(k)),
              ctypes.byref(c_dtype(alpha)),
              matdescra,
              val,
              indx,
              pntrb,
              pntre,
              x,
              ctypes.byref(c_dtype(beta)),
              y)

    @staticmethod
    def scsrmm(transa: np.char, m: int, n: int, k: int, alpha: np.single, matdescra: np.ndarray, val: np.ndarray,
               indx: np.ndarray, pntrb: np.ndarray, pntre: np.ndarray, b: np.ndarray, ldb: int, beta: np.single,
               c: np.ndarray, ldc: int):
        MklFunctions._csrmm(transa, m, n, k, alpha, matdescra, val, indx, pntrb, pntre, b, ldb, beta, c, ldc,
                            np.single)

    @staticmethod
    def dcsrmm(transa: np.char, m: int, n: int, k: int, alpha: np.double, matdescra: np.ndarray, val: np.ndarray,
               indx: np.ndarray, pntrb: np.ndarray, pntre: np.ndarray, b: np.ndarray, ldb: int, beta: np.double,
               c: np.ndarray, ldc: int):
        MklFunctions._csrmm(transa, m, n, k, alpha, matdescra, val, indx, pntrb, pntre, b, ldb, beta, c, ldc,
                            np.double)

    @staticmethod
    # Sparse matrix dense matrix eski
    def _csrmm(transa, m, n, k, alpha, matdescra, val, indx, pntrb, pntre, b, ldb, beta, c, ldc, dtype):
        c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

        def mkl_csrmm_argtypes():
            return [ctypes.c_char_p,
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(c_dtype),
                    ndpointer(dtype=ctypes.c_char, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=c_dtype, ndim=2, flags='C_CONTIGUOUS'),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(c_dtype),
                    ndpointer(dtype=c_dtype, ndim=2, flags='C_CONTIGUOUS'),
                    ctypes.POINTER(ctypes.c_int)]

        csrmm = mkl.mkl_scsrmm if dtype == np.single else mkl.mkl_dcsrmm

        csrmm.argtypes = mkl_csrmm_argtypes()

        csrmm(transa,
              ctypes.byref(ctypes.c_int(m)),
              ctypes.byref(ctypes.c_int(n)),
              ctypes.byref(ctypes.c_int(k)),
              ctypes.byref(c_dtype(alpha)),
              matdescra,
              val,
              indx,
              pntrb,
              pntre,
              b,
              ctypes.byref(ctypes.c_int(ldb)),
              ctypes.byref(c_dtype(beta)),
              c,
              ctypes.byref(ctypes.c_int(ldc)))

    @staticmethod
    def scsrmult(trans: np.char, request: int, sort: int, m: int, n: int, k: int, a: np.ndarray, ja: np.ndarray,
                 ia: np.ndarray, b: np.ndarray, jb: np.ndarray, ib: np.ndarray, c: np.ndarray, jc: np.ndarray,
                 ic: np.ndarray, nzmax: int, info: int):
        MklFunctions._csrmult(trans, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info, np.single)

    @staticmethod
    def dcsrmult(trans: np.char, request: int, sort: int, m: int, n: int, k: int, a: np.ndarray, ja: np.ndarray,
                 ia: np.ndarray, b: np.ndarray, jb: np.ndarray, ib: np.ndarray, c: np.ndarray, jc: np.ndarray,
                 ic: np.ndarray, nzmax: int, info: int):
        MklFunctions._csrmult(trans, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info, np.double)

    @staticmethod
    # Sparse matrix sparse matrix eski
    def _csrmult(trans, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info, dtype):
        c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

        def mkl_dcsrmult_argtypes():
            return [ctypes.c_char_p,
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.POINTER(ctypes.c_long),
                    ctypes.POINTER(ctypes.c_int)]

        csrmult = mkl.mkl_scsrmultcsr if dtype == np.single else mkl.mkl_dcsrmultcsr

        csrmult.argtypes = mkl_dcsrmult_argtypes()

        ja += 1
        ia += 1

        jb += 1
        ib += 1

        csrmult(trans,
                ctypes.byref(ctypes.c_int(request)),
                ctypes.byref(ctypes.c_int(sort)),
                ctypes.byref(ctypes.c_int(m)),
                ctypes.byref(ctypes.c_int(n)),
                ctypes.byref(ctypes.c_int(k)),
                a, ja, ia, b, jb, ib,
                c, jc, ic,
                ctypes.byref(ctypes.c_long(nzmax)),
                ctypes.byref(ctypes.c_int(info)))

        ja -= 1
        ia -= 1

        jb -= 1
        ib -= 1

        jc -= 1
        ic -= 1

    @staticmethod
    def scsradd(trans: np.char, request: int, sort: int, m: int, n: int, a: np.ndarray, ja: np.ndarray, ia: np.ndarray,
                beta: np.single, b: np.ndarray, jb: np.ndarray, ib: np.ndarray, c: np.ndarray, jc: np.ndarray,
                ic: np.ndarray, nzmax: int, info: int):
        MklFunctions._csradd(trans, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info, np.single)

    @staticmethod
    def dcsradd(trans: np.char, request: int, sort: int, m: int, n: int, a: np.ndarray, ja: np.ndarray, ia: np.ndarray,
                beta: np.double, b: np.ndarray, jb: np.ndarray, ib: np.ndarray, c: np.ndarray, jc: np.ndarray,
                ic: np.ndarray, nzmax: int, info: int):
        MklFunctions._csradd(trans, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info, np.double)

    @staticmethod
    def _csradd(trans, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info, dtype):
        c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

        def mkl_csradd_argtypes():
            return [ctypes.c_char_p,
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ctypes.POINTER(ctypes.c_int),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.POINTER(c_dtype),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=c_dtype, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ndpointer(dtype=ctypes.c_long, ndim=1, flags='C_CONTIGUOUS'),
                    ctypes.POINTER(ctypes.c_long),
                    ctypes.POINTER(ctypes.c_int)]

        csradd = mkl.mkl_scsradd if dtype == np.single else mkl.mkl_dcsradd

        csradd.argtypes = mkl_csradd_argtypes()

        ja += 1
        ia += 1

        jb += 1
        ib += 1

        csradd(trans,
               ctypes.byref(ctypes.c_int(request)),
               ctypes.byref(ctypes.c_int(sort)),
               ctypes.byref(ctypes.c_int(m)),
               ctypes.byref(ctypes.c_int(n)),
               a, ja, ia,
               ctypes.byref(c_dtype(beta)),
               b, jb, ib,
               c, jc, ic,
               ctypes.byref(ctypes.c_long(nzmax)),
               ctypes.byref(ctypes.c_int(info)))

        ja -= 1
        ia -= 1

        jb -= 1
        ib -= 1

        jc -= 1
        ic -= 1


# Sparse matrix dense matrix yeni
# transa, m, n, k, alpha, matdescra, val, indx, pntrb, pntre, b, ldb, beta, c, ldc, dtype
def _sparse_mm(operation, m, n, k, alpha, descr, val, indx, pntrb, pntre, layout, x, ldx, beta, y, ldy, dtype):
    c_dtype = ctypes.c_float if dtype == np.single else ctypes.c_double

    ref = MklFunctions._create_handle(enums.SparseIndexBaseT.SPARSE_INDEX_BASE_ZERO,
                                      m,
                                      n,
                                      pntrb,
                                      pntre,
                                      indx,
                                      val,
                                      dtype)

    MklFunctions._sp_optimize_mm(ref, operation, descr, layout, n, 1)

    sparse_mm = mkl.mkl_sparse_s_mm if dtype == np.single else mkl.mkl_sparse_d_mm

    sparse_mm.argtypes = [ctypes.c_int,
                          c_dtype,
                          MklFunctions.sparse_matrix_t,
                          enums.MatrixDescr,
                          ctypes.c_int,
                          ndpointer(dtype=c_dtype, ndim=2, flags='C_CONTIGUOUS'),
                          ctypes.c_int,
                          ctypes.c_int,
                          c_dtype,
                          ndpointer(dtype=c_dtype, ndim=2, flags='C_CONTIGUOUS'),
                          ctypes.c_int]

    sparse_mm(ctypes.c_int(operation.value),
              c_dtype(alpha),
              ref,
              descr,
              ctypes.c_int(layout.value),
              x,
              ctypes.c_int(n),
              ctypes.c_int(ldx),
              c_dtype(beta),
              y,
              ctypes.c_int(ldy))

    mkl.mkl_sparse_destroy(ref)



"""Includes LassoSelectionSettings class"""
from typing import Any, Dict

import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_selection.domain.objects.settings.base_supervised_feature_selection_settings import \
    BaseSupervisedFeatureSelectionSettings


class LassoSelectionSettings(BaseSupervisedFeatureSelectionSettings):
    """Settings for lasso selection service"""

    def __init__(self, data: pd.DataFrame, target: pd.DataFrame, target_type: TargetType = None,
                 lasso_args: Dict[str, Any] = None):
        super().__init__(data, target, target_type)
        self.lasso_args = lasso_args



"""Includes SweepSelectionSettings class"""
import pandas as pd

from organon.ml.feature_selection.domain.objects.settings.base_unsupervised_feature_selection_settings import \
    BaseUnsupervisedFeatureSelectionSettings


class SweepSelectionSettings(BaseUnsupervisedFeatureSelectionSettings):
    """Settings for sweep selection"""

    def __init__(self, data: pd.DataFrame, bin_count: int = None, r_factor: float = None, random_state: int = None,
                 max_col_count: int = None, n_threads: int = 1):
        super().__init__(data)
        self.bin_count = bin_count
        self.r_factor = r_factor
        self.random_state = random_state
        self.max_col_count = max_col_count
        self.n_threads = n_threads



"""Includes MultiLayerPerceptronRegressor class."""
from sklearn.neural_network import MLPRegressor

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_regressor import BaseSklearnRegressor


class MultiLayerPerceptronRegressor(BaseSklearnRegressor[MLPRegressor]):
    """Wrapper for sklearn MLPRegressor"""

    @property
    def regressor_class(self):
        return MLPRegressor



"""
This module includes AfePlotlyReportHelper class.
"""
from organon.afe.domain.reporting.base_afe_output_report import BaseAfeOutputReport
from organon.afe.domain.reporting.base_report_helper import BaseReportHelper
from organon.afe.domain.reporting.plotly_common_functions import PlotlyCommonFunctions


class AfePlotlyReportHelper(BaseReportHelper):
    """Class for reporting afe output using plotly"""

    @classmethod
    def generate_feature_extended_dashboard(cls, afe_report: BaseAfeOutputReport):
        raise NotImplementedError

    @classmethod
    def generate_feature_importance_dashboard(cls, afe_report: BaseAfeOutputReport, target_index: int = None,
                                              first_x_feature: int = None):
        raise NotImplementedError

    @classmethod
    def generate_roc_curve_dashboard(cls, afe_report: BaseAfeOutputReport, target_index: int = None, target_class=None):
        raise NotImplementedError

    @classmethod
    def generate_afe_report_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates afe report dashboard"""
        PlotlyCommonFunctions.generate_afe_report_dashboard(cls, afe_report)

    @classmethod
    def generate_afe_feature_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a table including afe output features information"""
        PlotlyCommonFunctions.generate_afe_feature_dashboard(cls, afe_report)

    @classmethod
    def generate_coarse_class_dashboard(cls, afe_report: BaseAfeOutputReport, feature_name: str):
        """Creates bar plot showing transformation map corresponding to given feature"""
        return PlotlyCommonFunctions.generate_coarse_class_dashboard(afe_report, feature_name)

    @classmethod
    def generate_memory_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plot of memory usage by date"""
        return PlotlyCommonFunctions.generate_memory_usage_dashboard(afe_report)

    @classmethod
    def generate_cpu_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plot of cpu usage by date"""
        return PlotlyCommonFunctions.generate_cpu_usage_dashboard(afe_report)

    @classmethod
    def generate_resource_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plots of memory and cpu usages by date"""
        PlotlyCommonFunctions.generate_resource_usage_dashboard(afe_report)



"""Class for Run Type Enum"""
from enum import Enum


class DqRunType(Enum):
    """Run type enum"""
    RUN_ONE_DATA_SOURCE = 1
    RUN_ONE_DATA_SOURCE_WITH_PARTITIONS = 2
    RUN_MULTIPLE_DATA_SOURCE = 3



"""Includes SingleSourceReportHelper class."""
from typing import List

import pandas as pd

from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.domain.reporting.base_dq_report_helper import BaseDqReportHelper
from organon.idq.domain.reporting.objects.dq_comparison_sample_report_input import DqComparisonSampleReportInput
from organon.idq.domain.reporting.objects.single_source_alert_info import SingleSourceAlertInfo
from organon.idq.domain.reporting.objects.single_source_report_helper_params import SingleSourceReportHelperParams
from organon.idq.domain.reporting.objects.single_source_report_output import SingleSourceOutputReport


class SingleSourceReportHelper(BaseDqReportHelper):
    """Includes helper methods for one df reporting DQ execution results"""

    def __init__(self, params: SingleSourceReportHelperParams):
        super().__init__(params)
        self.params: SingleSourceReportHelperParams = params

    def execute(self) -> SingleSourceOutputReport:
        """Execute reporting for one df"""
        report_data = self._get_report_data()
        alert_list = report_data.alert_info
        alerts_df = pd.DataFrame.from_records([a.to_dict() for a in alert_list])
        alerts_df_unique = alerts_df.drop_duplicates(
            subset=alerts_df.columns,
            keep='last').reset_index(drop=True)
        signal_list = SingleSourceReportHelper._get_signal_details(report_data)
        signals_df = pd.DataFrame.from_records([s.to_dict() for s in signal_list])
        output_report = SingleSourceOutputReport()
        output_report.alerts_df = alerts_df_unique
        output_report.signals_df = signals_df
        output_report.signals = report_data.dq_comparison_result
        output_report.column_collection = report_data.dq_data_column_collection
        return output_report

    def _get_report_data(self) -> DqComparisonSampleReportInput:

        report_data = DqComparisonSampleReportInput()
        report_data.dq_data_column_collection = self.data_column_collection
        report_data.dq_comparison_result = self.comparison_results
        alert_info_list = self._get_alert_data()
        report_data.alert_info = alert_info_list
        return report_data

    def _get_alert_data(self) -> List[SingleSourceAlertInfo]:

        list_base_alert_info: List[SingleSourceAlertInfo] = []
        for comparison_result in self.comparison_results:
            alert_info = SingleSourceAlertInfo()
            alert_info.data_source_name = self.data_source_name
            if comparison_result.data_entity == DataEntityType.TABLE:
                alert = DataEntityType.TABLE.name
            else:
                alert = comparison_result.data_entity_name
            alert_info.alert = alert
            alert_info.full_filter_str = self._get_full_filter_string(self.params.partition, self.params.filter_str)
            alert_info.run_type = self.run_type
            list_base_alert_info.append(alert_info)

        return list_base_alert_info



"""todo"""
from typing import Dict

from organon.idq.domain.businessobjects.statistics.dq_categorical_statistics import \
    DqCategoricalStatistics
from organon.idq.domain.businessobjects.statistics.dq_sample_numerical_statistics import \
    DqSampleNumericalStatistics


class SampleStatistics:
    """todo"""

    def __init__(self):
        self.numerical_statistics: Dict[str, DqSampleNumericalStatistics] = None
        self.nominal_statistics: Dict[str, DqCategoricalStatistics] = None



"""Includes SweepSelectionService class"""
from organon.ml.feature_selection.domain.objects.settings.sweep_selection_settings import SweepSelectionSettings
from organon.ml.feature_selection.domain.objects.sweep_selection_output import SweepSelectionOutput
from organon.ml.feature_selection.domain.services.base_feature_selection_service import SelectionRunSettings
from organon.ml.feature_selection.domain.services.base_unsupervised_feature_selection_service import \
    BaseUnsupervisedFeatureSelectionService


class SweepSelectionService(BaseUnsupervisedFeatureSelectionService[SweepSelectionSettings, SweepSelectionOutput]):
    """Service class for SweepSelection"""

    @classmethod
    def _validate_settings(cls, settings: SweepSelectionSettings):
        cls._validate_original_data_for_sweep(settings.data)

    @classmethod
    def _run(cls, run_settings: SelectionRunSettings) -> SweepSelectionOutput:
        settings: SweepSelectionSettings = run_settings.settings
        sweep_args = cls._get_sweep_args(settings)
        selected_features = cls._get_features_with_sweep(run_settings.data, sweep_args)
        output = SweepSelectionOutput()
        output.selected_features = selected_features
        return output

    @classmethod
    def _get_sweep_args(cls, settings: SweepSelectionSettings):
        sweep_args = {}
        if settings.bin_count is not None:
            sweep_args["bin_count"] = settings.bin_count
        if settings.r_factor is not None:
            sweep_args["r_factor"] = settings.r_factor
        if settings.random_state is not None:
            sweep_args["random_state"] = settings.random_state
        if settings.max_col_count is not None:
            sweep_args["max_col_count"] = settings.max_col_count
        if settings.n_threads is not None:
            sweep_args["n_threads"] = settings.n_threads
        return sweep_args



"""
This module includes IntervalType enum definition.
"""
from enum import Enum


class IntervalType(Enum):
    """IntervalType enum class"""
    OPEN_OPEN = 0
    OPEN_CLOSED = 1
    CLOSED_OPEN = 2
    CLOSED_CLOSED = 3



"""Includes UnsupervisedFeatureExtractor class."""
import random
import sys
from typing import Optional, List, Tuple, Dict

import numpy as np
from scipy.sparse import csc_matrix

from organon.fl.core.executionutil.objects.stopwatch import StopWatch
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.fl.mathematics.linearalgebra.blas_service_factory import BlasServiceFactory
from organon.fl.mathematics.sweep.covariance_generator import CovarianceGenerator
from organon.fl.mathematics.sweep.data_classes.transformation import Transformation
from organon.fl.mathematics.sweep.matrices.csc_matrix import CscMatrix
from organon.fl.mathematics.sweep.sweep_operator import UnsupervisedSweepOperator


class UnsupervisedFeatureExtractor:
    """
    Class for Unsupervised Feature Extraction Algorithm.
    """

    def __init__(self, data_frame: Dict[str, np.ndarray], bin_count: int = 4, r_factor: float = 0.9,
                 random_state: int = None, max_iter: int = 500,
                 is_logging: bool = False):
        self.__frame = data_frame
        self.__bin_count = bin_count
        self.__r_limit = bin_count * r_factor
        self.__random_state = random_state
        self.__max_iter = max_iter
        if max_iter < 0:
            self.__max_iter = sys.maxsize
        self.__is_logging = is_logging

        self.__un_swept = None
        self.__operator = None
        self.__generator = None
        self.__cov_info = None

    def log(self, log_str):
        """
        Logs the input str if logging is enables
        :param log_str: input str
        :return: None
        """
        if self.__is_logging:
            LogHelper.info(log_str)

    def execute(self, num_threads: int):
        """
            This is the main function.
            :param num_threads: number of parallel threads
            :return: None for now
            """
        watch = StopWatch()
        watch.start()
        input_frame, index_to_name = self.get_sparse_frame()

        if input_frame is None:
            return []

        self.log(f"Sparse Frame is Created in {watch.get_elapsed_seconds(True)}")

        random.seed(self.__random_state)
        first_enter = random.randrange(0, len(index_to_name))  # nosec
        self.log(index_to_name[first_enter])
        self.__un_swept = set(range(len(index_to_name)))

        sweep_operator, generator = self.get_sweep_operator_generator(input_frame)
        self.__operator = sweep_operator
        self.__generator = generator
        self.__cov_info = generator.covariance_info
        self.log(f"Sweep Operator and Cov Generator initialized in {watch.get_elapsed_seconds(True)}")

        self.include(first_enter, num_threads)

        for _ in range(self.__max_iter):
            if len(self.__un_swept) == 0:
                break

            next_swept = self.get_next_swept()

            if next_swept is None:
                break

            if next_swept[1] < self.__r_limit:
                break
            self.log(f'{index_to_name[next_swept[0]], next_swept[1]}')

            next_col = next_swept[0]
            self.include(next_col, num_threads)

        swept = set(range(len(index_to_name))).difference(self.__un_swept)
        # print(f'{swept=}')
        # print(f'{un_swept=}')
        return [index_to_name[index] for index in swept]

    def include(self, feature: int, num_threads: int):
        """
        Includes a feature via calculating correlation and sweeping.
        :param feature: Include feature
        :param num_threads: number of threads for parallel processing
        :return: None
        """
        bin_count = self.__bin_count
        self.__generator.compute_covariances_c(feature, num_threads, bin_count)
        start = feature * bin_count
        end = start + bin_count
        for col_extended in range(start, end):
            self.__operator.sweep(col_extended, num_threads)
        self.__un_swept.remove(feature)

    def get_next_swept(self) -> (int, float):
        """
        Gets next feature to sweep
        :return: index and value of a feature
        """
        bin_count = self.__bin_count
        next_swept: Optional[Tuple] = None

        for col in self.__un_swept:
            corr_tot = 0
            start = col * bin_count
            end = start + bin_count
            for i in range(start, end):
                corr_tot += self.__cov_info.correlation_matrix[i, i]

            if next_swept is None:
                next_swept = (col, corr_tot)
            elif corr_tot > next_swept[1]:  # pylint: disable=(unsubscriptable-object)
                next_swept = (col, corr_tot)

        return next_swept

    @staticmethod
    def get_sweep_operator_generator(sparse_frame: csc_matrix) -> (
            UnsupervisedSweepOperator, CovarianceGenerator):
        """
        Creates and initializes sweep operator and covariance generator for a sparse frame
        :param sparse_frame: input frame
        :return: sweep_operator, covariance_generator
        """
        rows, cols = sparse_frame.shape
        matrix = CscMatrix(sparse_frame.data, sparse_frame.indices, sparse_frame.indptr, shape=(rows, cols))

        transformations = []
        for _col in range(cols):
            transformations.append(Transformation(0, 0, 0, False))

        blas_service = BlasServiceFactory.get_blas_service()
        generator = CovarianceGenerator(False, matrix, transformations, np.ones(shape=(cols,)), blas_service)
        generator.initialize()
        cov_info = generator.covariance_info

        sweep_operator = UnsupervisedSweepOperator(cov_info.correlation_matrix, blas_service)
        generator.compute_variances()

        return sweep_operator, generator

    def split_bins(self, vector: np.ndarray) -> (np.ndarray, np.ndarray):
        """
        Splits a vector to bins with equal frequency.
        :param vector: input vector
        :return: sparse matrix
        """
        bin_count = self.__bin_count
        v_sorted_args = np.argsort(vector)
        elem_count = len(vector)
        dense = np.zeros(shape=(vector.shape[0], bin_count), dtype=vector.dtype)
        for i in range(bin_count):
            start = i * elem_count // (bin_count + 1)
            end = (i + 1) * elem_count // (bin_count + 1)
            dense[v_sorted_args[start:end], i] = 1
        ret_mtx = csc_matrix(dense)
        return ret_mtx.indices, ret_mtx.indptr

    def get_sparse_frame(self) -> Tuple[csc_matrix, List[str]]:
        """
        Gets binned sparse array from data_frame
        In order to concatenate sparse matrices fast. I used 3 array structure of csc matrix.
        :return: sparse matrix
        """
        arbitrary_column = next(iter(self.__frame.values()))
        bin_count = self.__bin_count

        n_r = len(arbitrary_column)
        n_c = len(self.__frame)

        nnz = n_r * n_c

        data = np.ones(shape=(nnz,), dtype=arbitrary_column.dtype)
        row_ptr = np.empty_like(data)
        col_ptr = np.zeros(shape=(n_c * bin_count + 1,), dtype=np.longlong)

        index_to_names = [''] * n_c

        row_start = 0
        index = 0
        for name, column in self.__frame.items():
            if np.isnan(column).all():
                continue
            row_ptr_vec, col_ptr_vec = self.split_bins(column)
            row_end = row_start + row_ptr_vec.shape[0]
            row_ptr[row_start:row_end] = row_ptr_vec
            row_start = row_end
            col_ptr[1 + index * bin_count:1 + (index + 1) * bin_count] = col_ptr[index * bin_count] + col_ptr_vec[1:]
            index_to_names[index] = name
            # print([np.sum(sparse_matrix[:, i]) for i in range(sparse_matrix.shape[1])])
            # del data_frame[name]
            index += 1
        if index == 0:
            return None, None
        nnz = col_ptr[index * bin_count]
        return (csc_matrix((data[:nnz], row_ptr[:nnz], col_ptr[:1 + index * bin_count]),
                           shape=(n_r, index * bin_count)), index_to_names[:index])



"""Includes SourceWithDateReportHelper class."""
from typing import List

from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.reporting.base_multiple_calculation_report_helper import BaseMultipleCalculationReportHelper
from organon.idq.domain.reporting.objects.source_with_partitions_alert_info import SourceWithPartitionsAlertInfo
from organon.idq.domain.reporting.objects.source_with_partitions_dq_report_output import \
    SourceWithPartitionsDqOutputReport
from organon.idq.domain.reporting.objects.source_with_partitions_report_helper_params import \
    SourceWithPartitionsReportHelperParams


class SourceWithPartitionsReportHelper(BaseMultipleCalculationReportHelper):
    """Includes helper methods for one df with date reporting DQ execution results"""

    def __init__(self, params: SourceWithPartitionsReportHelperParams):
        super().__init__(params)
        self.params = params

    def execute(self) -> SourceWithPartitionsDqOutputReport:
        """Execute reporting for one df"""

        report_data = self._get_report_data()
        output_report = SourceWithPartitionsDqOutputReport()
        self._fill_output_report(report_data, output_report)
        return output_report

    def _get_table_alert(self, table_comparison_result):
        table_alert_list = self._get_table_alert_list(self.calculation_count)
        alert_list = []

        if all(elem in table_comparison_result for elem in table_alert_list):
            alert_info = SourceWithPartitionsAlertInfo()
            alert_info.data_source_name = self.data_source_name
            alert_info.alert = DataEntityType.TABLE.name
            alert_info.run_type = self.run_type
            alert_info.full_filter_str = self._get_full_filter_string(self.params.partition,
                                                                      self.params.filter_str)
            alert_list.append(alert_info)
        return alert_list

    def _get_numeric_column_alert(self, numeric_comparison_result):
        numeric_alert_list = self._get_numeric_alert_list(self.calculation_count)
        column_stability_alert_list = self._get_column_stability_alert_list()

        return self._get_alert_list(numeric_comparison_result, numeric_alert_list, column_stability_alert_list)

    def _get_nominal_column_alert(self, nominal_comparison_result):
        nominal_alert_list = [DqComparisonResultCode.PSI_RED_SIGNAL]
        column_stability_alert_list = self._get_column_stability_alert_list()

        return self._get_alert_list(nominal_comparison_result, nominal_alert_list, column_stability_alert_list)

    def _get_alert_list(self, results: List[DqComparisonResult], alert_codes: List[DqComparisonResultCode],
                        column_stability_alert_list: List[DqComparisonResultCode]):
        column_results = self._get_result_dict(results)
        alert_list = []
        for column_name, result_code in column_results.items():
            if all(elem in result_code for elem in alert_codes) or \
                    any(elem in result_code for elem in column_stability_alert_list):
                alert_info = SourceWithPartitionsAlertInfo()
                alert_info.data_source_name = self.data_source_name
                alert_info.alert = column_name
                alert_info.run_type = self.run_type
                alert_info.full_filter_str = self._get_full_filter_string(self.params.partition, self.params.filter_str)
                alert_list.append(alert_info)
        return alert_list



"""Module for LinearRegressionStep"""
from dataclasses import dataclass

from organon.fl.mathematics.sweep.data_classes.linear_regression_results import LinearRegressionResults
from organon.fl.mathematics.sweep.data_classes.selected_attribute import SelectedAttribute


@dataclass
class LinearRegressionStep:
    """Step info for results"""
    step: int
    selected_attribute: SelectedAttribute
    anova: LinearRegressionResults

    def __str__(self) -> str:
        result_string = f"STEP NO: {self.step}\n"
        result_string += "\n"
        result_string += "SELECTED ATTRIBUTE INFORMATION:\n"
        result_string += str(self.selected_attribute)
        result_string += "\n"
        result_string += "ANALYSIS OF VARIANCE AT THIS STEP:\n"
        result_string += str(self.anova)

        return result_string



"""Module for SweptAttribute"""
from dataclasses import dataclass
import numpy as np


@dataclass
class SweptAttribute:
    """Info for swept attribute"""
    attribute_index: int = None
    unswept_row: np.ndarray = None
    unswept_col: np.ndarray = None
    swept_row: np.ndarray = None
    swept_col: np.ndarray = None



"""Includes BinaryObjectClassificationService"""
from typing import Tuple, Optional, TYPE_CHECKING

import numpy as np
import pandas as pd
from organon.ml.object_classification.common.object_classification_constants import ObjectClassificationConstants
from organon.ml.object_classification.domain.data_sources.tf_dataset_data_source_handler import \
    TfDatasetDataSourceHandler
from organon.ml.object_classification.domain.services.base_object_classification_service import \
    BaseObjectClassificationService

if TYPE_CHECKING:
    import tensorflow as tf  # noqa


class BinaryObjectClassificationService(BaseObjectClassificationService):
    """Class for Binary Object Classification"""

    def _create_final_dense_layer(self, base_model: "tf.Tensor", inputs: "tf.keras.Input",
                                  layer_num: int) -> "tf.keras.Model":
        import tensorflow as tf  # pylint: disable=import-outside-toplevel
        base_model = self._add_final_base_layers(base_model)
        outputs = self._get_output_layer(base_model, 1)
        return tf.keras.Model(inputs, outputs)

    def _get_output_layer(self, base_model: "tf.Tensor", layer_num: int) -> "tf.keras.layers.Dense":
        import tensorflow as tf  # pylint: disable=import-outside-toplevel
        return tf.keras.layers.Dense(layer_num)(base_model)

    def _get_compiled_model(self, fine_tuning: bool = False) -> "tf.keras.Model":
        import tensorflow as tf  # pylint: disable=import-outside-toplevel
        return self.model.compile(optimizer=self._get_optimizer(fine_tuning=fine_tuning),
                                  loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
                                  metrics=tf.keras.metrics.BinaryAccuracy())

    def _generate_image_data(self) -> Tuple[TfDatasetDataSourceHandler, TfDatasetDataSourceHandler]:
        return self._generate_train_validation_data(label_mode="binary")

    def _get_avg_parameter(self, average: Optional[str]) -> str:
        return "binary" if average is None else average

    def _get_predictions(self, probs: np.ndarray) -> np.ndarray:
        return np.where(probs > self._settings.prediction_threshold, 1, 0)

    def _set_probas(self, probs_df: pd.DataFrame, probs: np.ndarray) -> pd.DataFrame:
        probs_df.loc[:, ObjectClassificationConstants.PROBA_COL_NAME] = probs.flatten()
        return probs_df



""" This module includes BaseFeatureReduction"""
import abc
from typing import List
from pandas import DataFrame

from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.feature_reduction.domain.objects.feature_reduction_output import FeatureReductionOutput
from organon.ml.feature_reduction.settings.objects.base_feature_reduction_settings import BaseFeatureReductionSettings


class BaseFeatureReduction(metaclass=abc.ABCMeta):
    """ BaseFeatureReduction class"""

    def __init__(self, settings: BaseFeatureReductionSettings):
        self.settings = settings

    def execute(self) -> FeatureReductionOutput:
        """Execute Feature Reduction"""
        reduction_description = self.get_description()

        LogHelper.info(f"Executing reduction: {reduction_description}")
        results = self._execute_reduction(self.settings)

        LogHelper.info(f"Finished executing reduction: {reduction_description}")
        return results

    @staticmethod
    def _get_included_columns(data: DataFrame, excluded_columns: List[str]) -> List[str]:
        if excluded_columns is not None:
            included_columns = [col for col in data.columns if col not in excluded_columns]
        else:
            included_columns = list(data.columns)
        return included_columns

    @abc.abstractmethod
    def _execute_reduction(self, settings: BaseFeatureReductionSettings) -> FeatureReductionOutput:
        """
        Execute reductions and return reduction output.
        Parameters
        ----------
        settings

        Returns
        -------
        Return feature reduction output.
        """

    @abc.abstractmethod
    def get_description(self) -> str:
        """todo"""



"""Includes FeatureReductionOutput class"""
from typing import List

from organon.ml.feature_reduction.domain.enums.feature_reduction_types import FeatureReductionType


class FeatureReductionOutput:
    """Output of feature reduction"""

    def __init__(self):
        self.feature_reduction_type: FeatureReductionType = None
        self.reduced_column_list: List[str] = None



"""Includes DqComparisonReportInput class"""
from typing import List

from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.reporting.objects.base_alert_info import BaseAlertInfo
from organon.idq.domain.reporting.objects.nominal_column_control_details import NominalColumnControlDetails
from organon.idq.domain.reporting.objects.numeric_column_control_details import NumericColumnControlDetails
from organon.idq.domain.reporting.objects.table_control_details import TableControlDetails


class DqComparisonSampleReportInput:
    """Class for DqComparisonSampleReportInput"""

    def __init__(self):
        self.dq_data_column_collection: DqDataColumnCollection = None
        self.dq_comparison_result: List[DqComparisonResult] = None
        self.alert_info: List[BaseAlertInfo] = None
        self.numeric_columns_control: List[NumericColumnControlDetails] = None
        self.nominal_columns_control: List[NominalColumnControlDetails] = None
        self.table_control: List[TableControlDetails] = None



"""Includes RegressionReporterService class."""
from typing import List, Tuple

import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error, mean_squared_log_error
from sklearn.metrics import r2_score

from organon.ml.common.helpers.df_ops_helper import get_bins
from organon.ml.reporting.domain.objects.regression_report import RegressionReport
from organon.ml.reporting.domain.services.base_reporter_service import BaseReporterService
from organon.ml.reporting.settings.objects.reporter_settings import ReporterSettings


class RegressionReporterService(BaseReporterService[RegressionReport]):
    """Reporter service for regression"""

    def execute(self, settings: ReporterSettings) -> RegressionReport:
        id_str_exists = settings.id_str_column is not None
        split_col_exists = settings.split_column is not None
        group_by_columns = []
        if id_str_exists:
            group_by_columns.append(settings.id_str_column)
        if split_col_exists:
            group_by_columns.append(settings.split_column)

        lift_cols, performance_summary_cols = self._get_lift_and_perf_columns(id_str_exists, split_col_exists)
        if group_by_columns:
            performance_summary_list = []
            lift_list = []
            for groups, indices in settings.data.groupby(group_by_columns).indices.items():
                group_tuple = groups if isinstance(groups, tuple) else (groups,)
                performance_summary_list.append([*list(group_tuple), *self._calc_performance_summary_for_split(
                    settings.data.loc[indices], settings.target_column, settings.score_column).tolist()])
                lift_tables = self._create_lift_table_data(settings.data.loc[indices],
                                                           settings.target_column,
                                                           settings.score_column,
                                                           settings.num_bins)
                for i in range(settings.num_bins):
                    lift_list.append([*list(group_tuple), *lift_tables[i]])
            lift_table = pd.DataFrame(data=lift_list, columns=lift_cols)
            performance_summary_table = pd.DataFrame(data=performance_summary_list, columns=performance_summary_cols)
        else:
            lift_table = pd.DataFrame(data=self._create_lift_table_data(settings.data, settings.target_column,
                                                                        settings.score_column, settings.num_bins),
                                      columns=lift_cols)
            performance_summary_table = pd.DataFrame(data=[self._calc_performance_summary_for_split(
                settings.data, settings.target_column, settings.score_column).values], columns=performance_summary_cols)
        report = RegressionReport()
        report.performance_summary = performance_summary_table
        report.lift_table = lift_table
        return report

    @staticmethod
    def _get_lift_and_perf_columns(idstr_exists=False, split_exists=False) -> Tuple[List, List]:
        lift_columns = ["Bin", "Total Count", "Target Average", "Score Average", "Cumulative Lift"]
        perf_summary_columns = ["MSE", "RMSE", "MAA", "MAPE", "MSLOGE", "R2", "ROW COUNT", "TARGET AVG", "SCORE AVG",
                                "SCORE STD", "SCORE MIN", "SCORE P25", "SCORE MEDIAN", "SCORE P75", "SCORE MAX"]
        if split_exists:
            lift_columns.insert(0, "Data")
            perf_summary_columns.insert(0, "Data")
        if idstr_exists:
            lift_columns.insert(0, "IdStr")
            perf_summary_columns.insert(0, "IdStr")
        return lift_columns, perf_summary_columns

    def _get_empty_lift_perf_tables(self, idstr_exists=False, split_exists=False) -> Tuple[pd.DataFrame, pd.DataFrame]:
        return pd.DataFrame(columns=self._get_lift_and_perf_columns(idstr_exists, split_exists)[0]), \
               pd.DataFrame(columns=self._get_lift_and_perf_columns(idstr_exists, split_exists)[1])

    def _create_lift_table_data(self, split_df: pd.DataFrame, target_column: str,
                                score_column: str, num_bins: int) -> List:
        if num_bins > len(split_df):
            raise ValueError("num_bins cannot be higher than number of rows in a set")
        split_bins = get_bins(split_df[score_column], num_bins)
        return self._get_bin_list(split_df, target_column, split_bins, score_column, num_bins)

    @staticmethod
    def _get_bin_list(data: pd.DataFrame, target_column: str, bins: pd.Series, score_column: str,
                      num_bins: int) -> List:
        bin_list = []
        target_avg = data[target_column].mean()
        cumulative_sum, cumulative_count = 0, 0
        for bin_num in reversed(range(num_bins)):
            data_bins = data.loc[bins == bin_num]
            bin_target_avg = data_bins[target_column].mean()
            bin_length = len(data_bins)
            cumulative_sum += bin_target_avg * bin_length
            cumulative_count += bin_length
            bin_list.append([bin_num, bin_length, bin_target_avg,
                             data_bins[score_column].mean(),
                             cumulative_sum / (cumulative_count * target_avg)])
        bin_list.reverse()
        return bin_list

    def _calc_performance_summary_for_split(self, split_data: pd.DataFrame, target_col: str, prediction_col: str):
        df_statistics = split_data[prediction_col].agg(["mean", "std", "min", "max", "median"])
        percentiles = np.percentile(split_data[prediction_col], [25, 50, 75])
        return pd.Series(data=[mean_squared_error(split_data[target_col], split_data[prediction_col]),
                               mean_squared_error(split_data[target_col], split_data[prediction_col], squared=False),
                               mean_absolute_error(split_data[target_col], split_data[prediction_col]),
                               self._mean_absolute_percentage_error(split_data[target_col], split_data[prediction_col]),
                               mean_squared_log_error(split_data[target_col], split_data[prediction_col]),
                               r2_score(split_data[target_col], split_data[prediction_col]),
                               len(split_data),
                               split_data[target_col].mean(),
                               df_statistics["mean"],
                               df_statistics["std"],
                               df_statistics["min"],
                               percentiles[0],
                               percentiles[1],
                               percentiles[2],
                               df_statistics["max"]]
                         )

    @staticmethod
    def _mean_absolute_percentage_error(y_true, y_pred):
        y_true, y_pred = np.array(y_true), np.array(y_pred)
        y_true[y_true == 0] = 1
        return np.mean(np.abs((y_true - y_pred) / y_true)) * 100



"""Includes ModelCheckpoints enum class"""
from enum import Enum, auto


class ModelCheckpoints(Enum):
    """Pretrained model types"""

    ROBERTA_BASE = auto()
    DISTILBERT_BASE = auto()
    BERT_BASE = auto()
    BERT_BASE_TR_128 = auto()
    BERT_BASE_MLINGUAL = auto()
    BERT_BASE_TR_SENTIMENT = auto()



"""Includes NumericColumnControlDetails class"""

from organon.fl.core.enums.column_native_type import ColumnNativeType


class NumericColumnControlDetails:
    """Class for NumericColumnControlDetails """

    def __init__(self):
        self.calculation_name: str = None
        self.column_name: str = None
        self.column_type: ColumnNativeType = None
        self.mean: float = None
        self.trimmed_mean: float = None
        self.state_of_benchmark: int = None

    def to_dict(self):
        """return dict for base alert info"""
        return {
            'calculation_name':  self.calculation_name,
            'column_name': self.column_name,
            'column_type': self.column_type.name,
            'mean': self.mean,
            'trimmed_mean': self.trimmed_mean,
            'state_of_benchmark': self.state_of_benchmark
        }



"""Includes IDataPartition(base class for data partitioning classes) class."""
import abc
from typing import List

from organon.fl.core.businessobjects.idataframe import IDataFrame
from organon.fl.mathematics.businessobjects.tuple_generic_collection import TupleGenericCollection


class IDataPartition(metaclass=abc.ABCMeta):
    """Abstract base class for classes for data partitioning."""

    @property
    @abc.abstractmethod
    def frame(self) -> IDataFrame:
        """Returns the dataframe."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def training_ratio(self) -> float:
        """Returns the approximate value of 'size of training data/ size of all data'"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def training_indices(self) -> List[int]:
        """Indices of rows which will be used as training data."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def validation_indices(self) -> List[int]:
        """Indices of rows which will be used as validation data."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def strata_info(self) -> TupleGenericCollection[int]:
        """strata_info for partition"""
        raise NotImplementedError

    @abc.abstractmethod
    def partition(self, training_ratio: float):
        """Generates random training and validation indices according to training ratio."""
        raise NotImplementedError

    @abc.abstractmethod
    def set_frame(self, frame: IDataFrame):
        """Changes frame of the partition"""
        raise NotImplementedError



"""
This module includes TargetFileRecordRepository class.
"""
import math
from datetime import datetime
from typing import Callable, List, Optional

import numpy as np
import pandas as pd

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.dataaccess.helpers.record_source_helper import validate_entity_column
from organon.afe.dataaccess.services.base_record_reader import BaseRecordReader
from organon.afe.domain.common import afe_date_helper
from organon.afe.domain.enums.afe_date_column_type import AfeDateColumnType
from organon.afe.domain.enums.afe_target_column_type import AfeTargetColumnType
from organon.afe.domain.enums.binary_target_class import BinaryTargetClass
from organon.afe.domain.modelling.businessobjects.target_file_record import TargetFileRecord
from organon.afe.domain.modelling.businessobjects.target_file_record_collection import TargetFileRecordCollection
from organon.afe.domain.settings.record_source import RecordSource
from organon.afe.domain.settings.target_descriptor import TargetDescriptor
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import date_helper
from organon.fl.core.helpers.string_helper import is_null_or_empty


class TargetFileRecordRepository(BaseRecordReader):
    """
    This class is used to create a TargetFileRecordCollection, given a TargetFileDescriptor and record source.
    """

    def __init__(self, record_source: RecordSource, file_descriptor: TargetDescriptor,
                 rejected_target_classes: List[str] = None, is_scoring=False):
        super().__init__(record_source)
        if rejected_target_classes is None:
            rejected_target_classes = []
        self._record_source: RecordSource = record_source
        self._file_descriptor: TargetDescriptor = file_descriptor
        self._rejected_target_classes = rejected_target_classes
        self._is_scoring = is_scoring
        data_column_type: AfeDateColumnType = AfeDateColumnType.DateTime
        custom_format = None
        if self._file_descriptor.date_column is not None:
            data_column_type: AfeDateColumnType = self._file_descriptor.date_column.date_column_type
            custom_format: str = self._file_descriptor.date_column.custom_format
        self._func: Callable[[str], datetime] = \
            afe_date_helper.get_str_to_date_converter(data_column_type, custom_format)

    def read_text_source(self) -> TargetFileRecordCollection:
        """
        Reads text source.

        :return: TargetFileRecordCollection containing text records
        """
        with open(self._record_source.source, "r", encoding="utf8") as file:
            row_count = sum(1 for row in file if row.strip() != "") - 1  # subtract for header line

        collection = TargetFileRecordCollection(row_count)
        chunksize = self._file_descriptor.reading_settings.number_of_rows_per_step
        initial_index = 0
        for data_frame in pd.read_csv(self._record_source.source, sep=",", chunksize=chunksize):
            self.__get_records_from_dataframe(collection, data_frame, initial_index)  # pylint: disable=(no-member)
            initial_index += len(data_frame)
        return collection

    def read_dataframe_source(self) -> TargetFileRecordCollection:
        """
        Reads pandas DataFrame source.
        :return: TargetFileRecordCollection containing pandas DataFrame records
        """
        data_frame: pd.DataFrame = self._record_source.source
        collection = TargetFileRecordCollection(len(data_frame))
        return self.__get_records_from_dataframe(collection, data_frame)

    def read_db_source(self) -> TargetFileRecordCollection:
        raise NotImplementedError

    def __get_records_from_dataframe(self, collection: TargetFileRecordCollection, data_frame: pd.DataFrame,
                                     initial_index=0):
        """
        Creates TargetFileRecords from a pandas DataFrame and stores in collection.

        :param TargetFileRecordCollection collection:
        :param pd.DataFrame data_frame:
        :param int initial_index: collection index to start adding data_frame rows
        """
        index = initial_index
        validate_entity_column("target", data_frame, self._file_descriptor.entity_column_name)
        try:
            for row_iter in data_frame.iterrows():
                frame_row = row_iter[1]
                entity_id = frame_row[self._file_descriptor.entity_column_name]
                target = None
                if self._is_scoring:
                    if self._file_descriptor.target_column is not None and \
                            self._file_descriptor.target_column.column_name in frame_row:
                        target = frame_row[self._file_descriptor.target_column.column_name]
                else:
                    if self._file_descriptor.target_column is not None:
                        target = frame_row[self._file_descriptor.target_column.column_name]
                if entity_id is None or (isinstance(entity_id, float) and math.isnan(entity_id)):
                    continue
                if self._file_descriptor.date_column is not None:
                    date_value = frame_row[self._file_descriptor.date_column.column_name]
                    if date_value is None or (isinstance(date_value, float) and math.isnan(date_value)):
                        continue
                    record = self._get_record_from_row([entity_id, date_value, target])
                else:
                    record = self._get_record_from_row([entity_id, target])
                if record is not None:
                    collection.append(record)
                    index += 1
        except KeyError as key_err:
            raise KnownException(f"Column {key_err.args[0]} could not be found in data source.") from key_err
        return collection

    def _get_record_from_row(self, row, contains_target_column=True) -> Optional[TargetFileRecord]:
        """
        Creates a record from a Subscriptable storing entity_id in index 0, target value in index 1 and date in index 2

        :param row: A subscriptable object
        :return: TargetFileRecord created from the iterable 'row'
        """
        rejected_target_classes = self._rejected_target_classes
        binary_val = BinaryTargetClass.NAN
        double_val = np.float64("nan")
        multi_class_val = "nan"

        if self._file_descriptor.target_column is not None:
            binary_target_info = self._file_descriptor.target_column.binary_target_info

            if contains_target_column:
                target = self.__get_target(row)
                if self._file_descriptor.target_column.target_column_type == AfeTargetColumnType.Binary:
                    if target == binary_target_info.positive_category:
                        binary_val = BinaryTargetClass.POSITIVE
                    elif target == binary_target_info.negative_category:
                        binary_val = BinaryTargetClass.NEGATIVE
                    elif target == binary_target_info.indeterminate_category:
                        binary_val = BinaryTargetClass.INDETERMINATE
                    elif target == binary_target_info.exclusion_category:
                        binary_val = BinaryTargetClass.EXCLUSION
                    if rejected_target_classes is not None:
                        if target in rejected_target_classes:
                            return None
                elif self._file_descriptor.target_column.target_column_type == AfeTargetColumnType.Scalar:
                    double_val = np.float64(target)
                elif self._file_descriptor.target_column.target_column_type == AfeTargetColumnType.MultiClass:
                    multi_class_val = np.object_(target)
                else:
                    raise KnownException(
                        "Unknown target type : " + self._file_descriptor.target_column.target_column_type)

        date_val = self.__get_date_value(row)
        record = TargetFileRecord()
        record.entity_id = str(row[0])
        record.event_date = date_helper.get_date_as_integer(date_val)
        record.target_binary = binary_val
        record.target_scalar = double_val
        record.target_multi_class = multi_class_val

        return record

    def __get_target(self, row):
        if self._file_descriptor.date_column is not None:
            target = str(row[2])
        else:
            target = str(row[1])
        return target

    def __get_date_value(self, row):
        if self._file_descriptor.date_column is not None:
            date_val = self._file_descriptor.default_measurement_date
            if not is_null_or_empty(self._file_descriptor.date_column.column_name):
                date_as_string = str(row[1])
                date_val = self._func(date_as_string)
        else:
            date_val = AfeStaticObjects.no_date_col_default_date
        return date_val



"""Includes RandomFlipType enum class"""
from enum import Enum, auto


class RandomFlipType(Enum):
    """Random flip types"""
    VERTICAL = auto()
    HORIZONTAL = auto()
    HORIZONTAL_AND_VERTICAL = auto()
    


"""
This module includes AfeDataSettings class.
"""

from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.settings.base_afe_data_settings import BaseAfeDataSettings
from organon.afe.domain.settings.trx_descriptor import TrxDescriptor


class AfeDataSettings(BaseAfeDataSettings[TrxDescriptor]):
    """
    Stores data source settings for AfeModellingSettings
    """
    ATTR_DICT = {
        "trx_descriptor": TrxDescriptor,
    }
    ATTR_DICT.update(BaseAfeDataSettings.ATTR_DICT)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""
This module DateIntervalType enum class.
"""
from enum import Enum


class DateIntervalType(Enum):
    """
    Enum class for defining date interval type values.
    """
    OPEN_OPEN = 0
    OPEN_CLOSED = 1
    CLOSED_OPEN = 2
    CLOSED_CLOSED = 3



"""
This module includes SampleDataService for getting a sample data frame
"""

from datetime import datetime
from typing import Dict, Tuple

import pandas as pd

from organon.afe.dataaccess.services.i_sample_data_service import ISampleDataService
from organon.afe.dataaccess.services.sample_data_service_helper import SampleDataServiceHelper
from organon.afe.dataaccess.services.transaction_file_record_reader import TransactionFileRecordReader
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.base_afe_data_settings import BaseAfeDataSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.fl.core.businessobjects.date_interval import DateInterval


class SampleDataService(TransactionFileRecordReader, ISampleDataService):
    """
    This class handles getting sample data frame from different sources such as Database, dataframe, csv file.
    """

    def __init__(self, data_source_settings: BaseAfeDataSettings,
                 algorithm_settings: AfeAlgorithmSettings,
                 distinct_entity_source: RecordSource = None,
                 entity_source_entity_column: str = None,
                 interval_per_entity_per_date_column: Dict[str, Dict[str, DateInterval]] = None,
                 min_date: datetime = None,
                 max_date: datetime = None):
        # pylint: disable=too-many-arguments
        descriptor = data_source_settings.trx_descriptor
        trx_record_source = descriptor.modelling_raw_input_source

        date_col = descriptor.feature_gen_setting.date_column
        date_columns, min_max_dates_per_date_column = {}, {}
        if date_col is not None:
            date_columns = {date_col.column_name: date_col}
            if descriptor.feature_gen_setting.max_observation_date is not None:
                min_max_dates_per_date_column = {date_col.column_name:
                                                     (None, descriptor.feature_gen_setting.max_observation_date)}

        super().__init__(
            trx_record_source,
            date_columns,
            descriptor.entity_column_name,
            interval_per_entity_per_date_column=interval_per_entity_per_date_column,
            distinct_entity_source=distinct_entity_source,
            entity_source_entity_column=entity_source_entity_column,
            min_max_dates_per_date_column=min_max_dates_per_date_column,
            max_number_of_transaction_samples=data_source_settings.max_number_of_transaction_samples,
            dimension_compression_ratio=algorithm_settings.dimension_compression_ratio,
            dimension_max_cardinality=algorithm_settings.dimension_max_cardinality,
            reading_settings=data_source_settings.trx_descriptor.reading_settings
        )
        self.data_source_settings = data_source_settings
        self.__trx_entity_column: str = descriptor.entity_column_name
        self.__min_date = min_date
        self.__max_date = max_date
        self.helper = SampleDataServiceHelper(trx_record_source, self.__trx_entity_column,
                                              self.func_per_date_column, self.interval_per_entity_per_date_column,
                                              self.distinct_entity_source,
                                              self.data_source_settings.auto_column_decider_settings)

    def get_sample(self) -> pd.DataFrame:
        return self.read()

    def read_db_source(self):
        """
        Reads database source and creates a transaction record collection.
        :return: transaction record collection
        """
        raise NotImplementedError

    def read_dataframe_source(self):
        """
        Reads pandas DataFrame source.
        :return: TransactionFileRecordCollection containing DataFrame records
        """
        min_max_dates_per_date_column = self._get_intervals_per_date_column()
        unique_entities_set = None
        if self.distinct_entity_source is not None:
            unique_entities_set = self._get_unique_entities_set()

        return self.helper.read_dataframe_source(unique_entities_set, min_max_dates_per_date_column)

    def read_text_source(self):
        """
        Reads text source.
        :return: TransactionFileRecordCollection containing text records
        """
        min_max_dates_per_date_column = self._get_intervals_per_date_column()
        chunksize = self.data_source_settings.trx_descriptor.reading_settings \
            .number_of_rows_per_step
        unique_entities_set: set = None
        if self.distinct_entity_source is not None:
            unique_entities_set = self._get_unique_entities_set()
        return self.helper.read_text_source(unique_entities_set, min_max_dates_per_date_column, chunksize)

    def _get_intervals_per_date_column(self) -> Dict[str, Tuple[datetime, datetime]]:
        if self.__min_date is not None and self.__max_date is not None:
            if self.data_source_settings.trx_descriptor.feature_gen_setting.date_column is not None:

                return {self.data_source_settings.trx_descriptor.feature_gen_setting
                        .date_column.column_name: (self.__min_date,
                                                   self.__max_date)}
            return {}
        return super()._get_intervals_per_date_column()



"""
This module includes TransactionFileRecordCollection class.
"""
from typing import Dict

import numpy as np

from organon.afe.domain.modelling.businessobjects.transaction_file_stats import TransactionFileStats


class TransactionFileRecordCollection:
    """
    Collection of transaction file records
    """

    # pylint: disable=unused-argument
    def __init__(self, date_col_map: Dict[str, int], d_map: Dict[str, int], q_map: Dict[str, int],
                 max_record_count: int):
        super().__init__()
        self.transaction_file_stats: TransactionFileStats = None
        self.__date_col_map: Dict[str, int] = date_col_map
        self.__d_map: Dict[str, int] = d_map
        self.__q_map: Dict[str, int] = q_map
        self.__entity_index_map: Dict[str, np.array] = {}
        self.__dates: np.ndarray = np.full([max_record_count, len(date_col_map)], -1, dtype=np.int64)
        self.__q_arrays: np.ndarray = np.full([max_record_count, len(q_map)], None, dtype=np.float32)
        self.__d_arrays: np.ndarray = np.full([max_record_count, len(d_map)], -1, dtype=np.int8)

        self.__last_filled_index: int = None
        self.max_record_count = max_record_count

    @property
    def actual_record_count(self) -> int:
        """Number of rows that was filled with data.(Ignoring rows with dummy data)"""
        if self.__last_filled_index is None:
            return 0
        return self.__last_filled_index + 1

    @property
    def entity_index_map(self) -> Dict[str, np.array]:
        """Map of indices corresponding to given entity'"""
        return self.__entity_index_map

    @property
    def dates(self) -> np.ndarray:
        """Dates in transaction file records"""
        return self.__dates

    @property
    def d_arrays(self) -> np.ndarray:
        """Dimension values in transaction file records"""
        return self.__d_arrays

    @property
    def q_arrays(self) -> np.ndarray:
        """Quantity values in transaction file records"""
        return self.__q_arrays

    @property
    def date_col_map(self) -> Dict[str, int]:
        """Returns value of private attribute '__date_col_map'"""
        return self.__date_col_map

    @property
    def d_map(self) -> Dict[str, int]:
        """Returns value of private attribute '__d_map'"""
        return self.__d_map

    @property
    def q_map(self) -> Dict[str, int]:
        """Returns value of private attribute '__q_map'"""
        return self.__q_map

    def sort(self, date_column_name: str):
        """Sorts every list of transaction file records"""
        date_col_index = self.__date_col_map[date_column_name]
        new_entity_index_map = {}
        for entity, indices in self.__entity_index_map.items():
            indices_list: list = indices.tolist()
            new_entity_index_map[entity] = np.array(sorted(indices_list,
                                                           key=lambda _index: self.__dates[_index][date_col_index],
                                                           reverse=True), dtype=np.int32)
        self.__entity_index_map = new_entity_index_map

    def append(self, entity_id: str, trx_date_array: list, q_array: list, d_array: list):
        """Appends given transaction file record information to collection"""
        if self.__last_filled_index is None:
            self.__last_filled_index = -1
        self.__last_filled_index += 1
        index = self.__last_filled_index

        self.__dates[index] = trx_date_array
        self.__q_arrays[index] = q_array
        self.__d_arrays[index] = d_array

        if entity_id in self.__entity_index_map:  # bu kontrol her zaman olmalı
            self.__entity_index_map[entity_id].append(index)
        else:
            self.__entity_index_map[entity_id] = [index]

    def convert_entity_index_lists_to_array(self):
        """Converts entity index map values to numpy arrays"""
        for entity, index_list in self.__entity_index_map.items():
            if isinstance(index_list, list):
                self.__entity_index_map[entity] = np.array(index_list, dtype=np.int32)
